(message_name (identifier) @definition.class)

(enum_name (identifier) @definition.class)

(service_name (identifier) @definition.class)

(rpc_name (identifier) @definition.function)

(enum_variant_name (identifier) @definition.constant)

(field_name (identifier) @definition.constant)
