(comment) @comment.inside
(comment)+ @comment.around


[
  (disable_directive)
  (enable_directive)
  (extended_analysis_directive)
  (external_sources_directive)
  (source_directive)
  (source_path_directive)
  (shell_directive)
] @entry.around
