[
  (element)
  (start_tag)
  (if_statement)
  (else_if_block)
  (else_block)
  (then_block)
  (each_statement)
  (key_statement)
  (snippet_statement)
  (await_statement)
  (style_element)
  (script_element)
  (expression)
] @indent

[
  (end_tag)
  (if_end)
  (each_end)
  (await_end)
  (key_end)
  (snippet_end)
] @outdent
