; inherits: html

(raw_text) @none

[
  "as"
  "key"
  "html"
  "debug"
  "snippet"
  "render"
] @keyword

"const" @keyword.storage.modifier

[
  "if"
  "else if"
  "else"
  "then"
  "await"
] @keyword.control.conditional

"each" @keyword.control.repeat

"catch" @keyword.control.exception

[
  "{"
  "}"
] @punctuation.bracket

[
  "#"
  ":"
  "/"
  "@"
] @punctuation.delimiter
