;; def
(function_definition name: (symbol) @name) @definition.function

;; ext
(external_definition name: (symbol) @name) @definition.function

;; mac
(macro_definition name: (symbol) @name) @definition.function
