(function_definition) @local.scope
(constructor_definition) @local.scope
(block_statement) @local.scope

(function_definition (parameter name: (identifier) @local.definition.variable.parameter))
(constructor_definition (parameter name: (identifier) @local.definition.variable.parameter))

(identifier) @local.reference
