[
  (notes)
  (class_declaration)
  (inheritance)
  (feature_adaptation)
  (creation_clause)
  (converters)

  (feature_declaration)

  (attribute_or_routine)

  (precondition)
  (local_declarations)
  (feature_body)

  (check)
  (multi_branch)
  (multi_branch_expression)
  (conditional)
  (conditional_expression)
  (loop)
  (quantifier_loop)
  (iteration)

  (postcondition)
  (rescue)

  (invariant)
] @indent
