
((text_part) @injection.content
 (#set! injection.language "rst")
 (#set! injection.include-children))

((single_line_text_part) @injection.content
 (#set! injection.language "rst")
 (#set! injection.include-children))

