[
  (declaration_list)
  (compound_statement)
  (array_creation_expression)
  (subscript_expression)
  (parenthesized_expression)
  (formal_parameters)
  (arguments)
  (catch_clause)
  (foreach_statement)
  (switch_block)
  (anonymous_function_use_clause)
] @rainbow.scope

["(" ")" "[" "]" "{" "}"] @rainbow.bracket
