[
  (future_import_statement)
  (import_from_statement)
  (with_clause)
  (parameters)
  (parenthesized_list_splat)
  (argument_list)
  (tuple_pattern)
  (list_pattern)
  (subscript)
  (list)
  (set)
  (tuple)
  (dictionary)
  (dictionary_comprehension)
  (set_comprehension)
  (list_comprehension)
  (generator_expression)
  (parenthesized_expression)
  (interpolation)
  (format_expression)
] @rainbow.scope

[
  "(" ")"
  "{" "}"
  "[" "]"
] @rainbow.bracket

; (string ["{{" "}}"] @rainbow.bracket) @rainbow.scope
