[
  ; ()
  (arguments "(" ")")
  (parenthesized_expression)
  (function_type)
  ; #{}
  (record)
  (map)
  ; {}
  (map_update)
  (tuple)
  ; <<>>
  (bitstring)
  ; []
  (list)
] @rainbow.scope

[
  "#"
  "{" "}"
  "(" ")"
  "[" "]"
  "<<" ">>"
] @rainbow.bracket
