[
  (doctype)
  (erroneous_end_tag)
] @rainbow.scope

([
   (element)
   (script_element)
   (style_element)
 ] @rainbow.scope
 (#set! rainbow.include-children))

["<" ">" "<!" "</" "/>"] @rainbow.bracket
