(signature name: (variable) @local.definition.function)
(function name: (variable) @local.definition.function)
(pattern/variable) @local.definition.variable.parameter
(expression/variable) @local.reference
