[
  (par_arr_lit)
  (par_tup_lit)
  (sqr_arr_lit)
  (sqr_tup_lit)
  (tbl_lit)
  (struct_lit)
  (short_fn_lit)
] @rainbow.scope

[
  "(" "@(" ")"
  "[" "@[" "]"
  "{" "@{" "}"
  "|"
] @rainbow.bracket
