(functional_notation
  (atom)
  (open_ct) @indent
  (close) @outdent)

(list_notation
  (open_list) @indent
  (close_list) @outdent)

(curly_bracketed_notation
  (open_curly) @indent
  (close_curly) @outdent)
