[
  (keyframe_block_list)
  (block)
  (attribute_selector)
  (feature_query)
  (parenthesized_query)
  (selector_query)
  (parenthesized_value)
  (arguments)
] @rainbow.scope

[
  "{" "}"
  "(" ")"
] @rainbow.bracket
