[
  (cast_expression)
  (inferred_parameters)
  (dimensions_expr)
  (parenthesized_expression)
  (array_access)
  (argument_list)
  (type_arguments)
  (dimensions)
  (block)
  (switch_block)
  (catch_clause)
  (resource_specification)
  (for_statement)
  (enhanced_for_statement)
  (annotation_argument_list)
  (element_value_array_initializer)
  (module_body)
  (enum_body)
  (type_parameters)
  (class_body)
  (constructor_body)
  (annotation_type_body)
  (annotation_type_element_declaration)
  (interface_body)
  (array_initializer)
  (formal_parameters)
] @rainbow.scope

[
  "(" ")"
  "{" "}"
  "[" "]"
  "<" ">"
] @rainbow.bracket
