(text) @string
(filepath) @string.special.path

(change type: "A" @diff.plus)
(change type: "D" @diff.minus)
(change type: "M" @diff.delta)
(change type: "C" @diff.plus)
(change type: "R" @diff.delta)

(comment) @comment
