[
  (function_definition)
  (compound_statement)
  (subshell)
  (test_command)
  (subscript)
  (parenthesized_expression)
  (array)
  (expansion)
  (command_substitution)
] @rainbow.scope

[
  "(" ")"
  "((" "))"
  "${" "$("
  "{" "}"
  "[" "]"
  "[[" "]]"
] @rainbow.bracket
