[
  (block)
  (impl)
  (attribute)
  (generic_parameters)
  (parameter_list)
  (impl)
  (trait)
  (enum_body)
  (enum_member)
  (mod)
  (reg_reset)
  (reg_initial)
  (match_block)
  (lambda_params)
  (paren_expression)
  (turbofish)
  (stage_reference)
  (argument_list)
  (braced_parameter_list)
  (tuple_literal)
  (array_type)
  (array_literal)
  (array_pattern)
  (index)
  (tuple_pattern)
] @rainbow.scope


[
  "[" "]"
  "(" ")"
  "{" "}"
  "<" ">"
  "$("
  "::<"
  "::$<"
] @rainbow.bracket
