[
  (symbol)
  (string)
] @local.reference

[
  (config)
  (menuconfig)
  (choice)
  (comment_entry)
  (menu)
  (if)
] @local.scope

(type_definition (string) @local.definition.type)
(type_definition (input_prompt (string) @local.definition.type))
(type_definition_default (expression (string) @local.definition.type))
