[
  (param_list)
  (array)
  (block)
  (call)
  (class_def)
  (struct_def)
  (module_def)
  (enum_def)
  (lib_def)
  (fun_def)
  (method_def)
  (case)
  (select)
  (elsif)
  (if)
  (unless)
  (hash)
  (tuple)
] @indent

[
  ")"
  "}"
  "]"
  "end"
  "when"
  "in"
] @outdent
