; Scopes

[
  (global_decl)
  (switch_body)
  (compound_statement)
] @local.scope

; Definitions

(param
  (identifier) @local.definition.variable.parameter)

; References

(identifier) @local.reference
; (type_specifier) @local.reference

