; Folds for code blocks
[
  (statement_block)
  (component_body)
  (class_body)
  (object)
  (object_pattern)
  (array)
  (array_pattern)
  (switch_body)
] @fold

; Fold multi-line JSX elements
(jsx_element) @fold

; Fold style elements
(style_element) @fold

; Fold server blocks
(server_block) @fold

; Fold comments
(comment) @fold

; Fold template strings
(template_string) @fold
