((interpolation) @injection.content
 (#set! injection.language "haxe"))
((comment) @injection.content
 (#set! injection.language "jsdoc"))
; ((comment) @injection.content
;  (#set! injection.language "comment"))
