[
  (module_definition)
  (struct_definition)
  (macro_definition)
  (function_definition)
  (if_statement)
  (try_statement)
  (for_statement)
  (while_statement)
  (let_statement)
  (quote_statement)
  (do_clause)
  (compound_statement) ; begin block
] @fold
