[
  (begin_block)
  (end_block)
  (singleton_method)
  (block_parameters)
  (parenthesized_statements)
  (element_reference)
  (argument_list "(" ")")
  (block)
  (destructured_left_assignment)
  (interpolation)
  (string_array)
  (symbol_array)
  (regex)
  (array)
  (hash)
  (method_parameters)
] @rainbow.scope

[
  "#{"
  "{" "}"
  "(" ")"
  "%w(" "%i("
  "[" "]"
  "|"
  "/"
] @rainbow.bracket
