[
  (preproc_params)
  (preproc_defined)
  (argument_list)
  (attribute_specifier)
  (ms_declspec_modifier)
  (declaration_list)
  (parenthesized_declarator)
  (parenthesized_expression)
  (abstract_parenthesized_declarator)
  (array_declarator)
  (compound_statement)
  (initializer_list)
  (compound_literal_expression)
  (enumerator_list)
  (field_declaration_list)
  (parameter_list)
  (for_statement)
  (macro_type_specifier)
  (subscript_expression)
  (subscript_designator)
  (cast_expression)
] @rainbow.scope

[
  "(" ")"
  "{" "}"
  "[" "]"
] @rainbow.bracket
