(class
  name: _ @definition.struct)

(trait
  name: _ @definition.interface)

(external_function
  name: _ @definition.function)

(method
  name: _ @definition.function)

(define_constant
  name: _ @definition.constant)
