[
  (target_group)
  (unqualified_imports)
  (tuple)
  (list)
  (function)
  (function_parameters)
  (todo)
  (tuple)
  (list)
  (anonymous_function)
  (block)
  (case)
  (record_update)
  (arguments)
  (record_pattern_arguments)
  (tuple_pattern)
  (list_pattern)
  (type_definition)
  (data_constructor_arguments)
  (tuple_type)
  (function_parameter_types)
  (type_arguments)
  (type_parameters)
] @rainbow.scope

[
  "(" ")"
  "[" "]"
  "{" "}"
  "#"
] @rainbow.bracket
