[
  (formals)
  (parenthesized_expression)
  (attrset_expression)
  (let_attrset_expression)
  (rec_attrset_expression)
  (inherit_from)
  (interpolation)
  (list_expression)
] @rainbow.scope

[
  "${"
  "{" "}"
  "(" ")"
  "[" "]"
] @rainbow.bracket
