; inherits: html

(directive) @keyword.directive
(directive_start) @keyword.directive
(directive_end) @keyword.directive
(comment) @comment

; merged with blade punctuation
[
  "{{"
  "}}"
  "{!!"
  "!!}"
  "("
  ")"
] @punctuation.bracket
