[
  (for_statement)
  (if_statement)
  (while_statement)
  (do_statement)
  (function_declaration)
  (sub_declaration)
] @indent

[
  "NEXT"
  "END IF"
  "WEND"
  "LOOP"
  "END FUNCTION"
  "END SUB"
] @outdent
